#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Steganography-Tools - A Python steganography module.
# Copyright (C) 2021 Sébastien Mascha, Thomas Le Couedic
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

__author__ = "Sebastien Mascha, Thomas Le Couedic"
__version__ = "$Revision: 0.1 $"
__date__ = "$Date: 2021/04/01 $"
__revision__ = "$Date: 2021/04/01 $"
__license__ = "GPLv3"


import numpy as np
import matplotlib.pyplot as plt
import scipy
from scipy import ndimage
import cv2
import imageio
from PIL import Image, ExifTags

def img_information(img):
    "This function allows us to get the basic information of an image."
    # Opening Image as an object
    image = Image.open(img)
    # Getting the filename of image
    print("Filename : ",image.filename)
    # Getting the format of image
    print("Format : ",image.format)
    # Getting the mode of image
    print("Mode : ",image.mode)
    # Getting the size of image
    print("Size : ",image.size)
    # Getting only the width of image
    print("Width : ",image.width)
    # Getting only the height of image
    print("Height : ",image.height)
    # Getting the color palette of image
    print("Image Palette : ",image.palette)
    print("\n")
    # Showing image
    image.show()
    # Closing Image object
    image.close()
    
def get_metadata(img):
    "Check for potential metadata."
    image = Image.open(img)
    exif = { ExifTags.TAGS[k]: v for k, v in image._getexif().items() if k in ExifTags.TAGS }
    return exif

def rgb2grayscale(img):
    "Convert RGB image to grayscale using PIL"
    image = Image.open(img)
    image = image.convert('L')
    image.show()
    image.save('image_grayscale.jpg')
    
def plot_histogram(img):
    "Plot pixels distribution"
    fig = plt.figure(figsize=(8, 6))
    plt.hist(img.ravel(),bins = 256, range = [0,256]) 
    plt.title('Histogram representing pixel distribution', fontsize=25)
    plt.xlabel('Pixel value', fontsize=20)
    plt.ylabel('Count', fontsize=20)
    plt.show()
    
def thresholding(img, threshold=65):
    "Black or white pixels according to a given threshold"
    img = imageio.imread(img) 
    img = img[:,:,0]
    for x in range(img.shape[0]):
        for y in range(img.shape[1]):
            if img[x,y] > threshold :
                img[x,y] = 255
            else:
                img[x,y] = 0
    
    fig = plt.figure(figsize=(8, 6))
    plt.imshow(img, cmap=plt.cm.gray)
    plt.title('Image displayed with a threshold for the pixels', fontsize=25)
    plt.show
    
def display_images(img):
    "Display images with different contrasts"
    fig = plt.figure(figsize=(24,20))

    ax1 = fig.add_subplot(131)
    ax1.title.set_text('Original image')
    plt.imshow(img, cmap=plt.cm.gray)
    
    ax2 = fig.add_subplot(132)
    ax2.title.set_text("Contrasted image : dark")
    plt.imshow(img, cmap=plt.cm.gray, vmin=10, vmax=500)
    plt.axis('off')

    ax3 = fig.add_subplot(133)
    ax3.title.set_text("Contrasted image : lightened")
    plt.imshow(img, cmap=plt.cm.gray, vmin=10, vmax=80)
    plt.axis('off')

    plt.show()

def image_manipulation(img):
    "Draw circles, lines into the image"
    img = imageio.imread(img) 
    img = img[:,:,0]
    
    img[100:120] = 255

    lx, ly = img.shape
    X, Y = np.ogrid[0:lx, 0:ly]
    mask = (X - lx / 2)**2 + (Y - ly / 2)**2 > (lx / 2)**2
    img[mask] = 0

    fig = plt.figure(figsize=(10, 8))
    plt.imshow(img, cmap=plt.cm.gray)
    plt.title('Image manipulated with Numpy', fontsize=25)
    plt.axis('off')

    plt.show()
    
    
def geo_transfomation(img):
    "Data augmentation/Geometrical trasnformation : cropping, up flip, down flip, rotations"
    lx, ly = img.shape
    crop_img = img[lx//4:-lx//4, ly//4:-ly//4]
    flip_ud_img = np.flipud(img)
    rotate_img = ndimage.rotate(img, 45)
    rotate_img_noreshape = ndimage.rotate(img, 45, reshape=False)

    fig = plt.figure(figsize=(12.5, 2.5))

    ax1 = fig.add_subplot(151)
    ax1.title.set_text('Original image')
    plt.imshow(img, cmap=plt.cm.gray)
    plt.axis('off')
    ax2 = fig.add_subplot(152)
    ax2.title.set_text('Image cropped')
    plt.imshow(crop_img, cmap=plt.cm.gray)
    plt.axis('off')
    ax3 = fig.add_subplot(153)
    ax3.title.set_text('Image flipped')
    plt.imshow(flip_ud_img, cmap=plt.cm.gray)
    plt.axis('off')
    ax4 = fig.add_subplot(154)
    ax4.title.set_text("Image rotated')
    plt.imshow(rotate_img, cmap=plt.cm.gray)
    plt.axis('off')
    ax5 = fig.add_subplot(155)
    ax5.title.set_text("Image rotated noreshape')
    plt.imshow(rotate_img_noreshape, cmap=plt.cm.gray)
    plt.axis('off')

    plt.subplots_adjust(wspace=0.02, hspace=0.3, top=1, bottom=0.1, left=0,
                    right=1)

    plt.show()
    
def blurring(img):
    "Use gaussian filers for blurring"
    blurred_img = ndimage.gaussian_filter(img, sigma=3)
    very_blurred = ndimage.gaussian_filter(img, sigma=9)
    local_mean = ndimage.uniform_filter(img, size=11)

    fig = plt.figure(figsize=(15,5))
    ax1 = fig.add_subplot(131)
    ax1.title.set_text('Blurred image - Gaussian filter =3')
    plt.imshow(blurred_img, cmap=plt.cm.gray)
    plt.axis('off')
    ax2 = fig.add_subplot(132)
    ax2.title.set_text('Very blurred image - Gaussian filter =9')
    plt.imshow(very_blurred, cmap=plt.cm.gray)
    plt.axis('off')
    ax3= fig.add_subplot(133)
    ax3.title.set_text('Image with uniform filter')             
    plt.imshow(local_mean, cmap=plt.cm.gray)
    plt.axis('off')

    plt.subplots_adjust(wspace=0.1, hspace=0., top=0.99, bottom=0.01,
                    left=0.01, right=0.99)

    plt.show()

def sharpenning(img,alpha=30):
    "Sharpenning"
    blurred_img = ndimage.gaussian_filter(img, 3)

    filter_blurred_img = ndimage.gaussian_filter(blurred_img, 1)

    sharpened = blurred_img + alpha * (blurred_img - filter_blurred_img)

    fig = plt.figure(figsize=(18, 6))
 
    ax1 = fig.add_subplot(131)
    ax1.title.set_text('Original image')
    plt.imshow(img, cmap=plt.cm.gray)
    plt.axis('off')
    ax2 = fig.add_subplot(132)
    ax2.title.set_text('Blurred image - Gaussian filter =3')
    plt.imshow(blurred_img, cmap=plt.cm.gray)
    plt.axis('off')
    ax3 = fig.add_subplot(133)
    ax3.title.set_text('Blurred image sharpenned')
    plt.imshow(sharpened, cmap=plt.cm.gray)
    plt.axis('off')

    plt.tight_layout()
    plt.show()
    
def denoising(img):
    "Denoising"
    noisy = img + 0.4*img.std()*np.random.random(img.shape)

    gauss_denoised = ndimage.gaussian_filter(noisy, 2)
    med_denoised = ndimage.median_filter(noisy, 3)

    plt.figure(figsize=(18,6))

    plt.subplot(131)
    plt.imshow(noisy, cmap=plt.cm.gray, vmin=40, vmax=220)
    plt.axis('off')
    plt.title('noisy', fontsize=20)
    plt.subplot(132)
    plt.imshow(gauss_denoised, cmap=plt.cm.gray, vmin=40, vmax=220)
    plt.axis('off')
    plt.title('Gaussian filter', fontsize=20)
    plt.subplot(133)
    plt.imshow(med_denoised, cmap=plt.cm.gray, vmin=40, vmax=220)
    plt.axis('off')
    plt.title('Median filter', fontsize=20)

    plt.subplots_adjust(wspace=0.02, hspace=0.02, top=0.9, bottom=0, left=0,
                        right=1)
    plt.show()
    


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Steganography-Tools - A Python steganography module.
# Copyright (C) 2021 Sébastien Mascha, Thomas Le Couedic
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

__author__ = "Sebastien Mascha, Thomas Le Couedic"
__version__ = "$Revision: 0.1 $"
__date__ = "$Date: 2021/04/01 $"
__revision__ = "$Date: 2021/04/01 $"
__license__ = "GPLv3"

import math

import numpy as np


class LSBSteganographyException(Exception):
    pass


class LSBSteganography():
    """
    Main class to handle Steganography encrypted data.
    """

    def __init__(self, im):
        self.image = im
        self.height, self.width, self.nbchannels = im.shape
        self.size = self.width * self.height
        
        self.maskONEValues = [1,2,4,8,16,32,64,128]
        #Mask used to put one ex:1->00000001, 2->00000010 .. associated with OR bitwise
        self.maskONE = self.maskONEValues.pop(0) #Will be used to do bitwise operations
        
        self.maskZEROValues = [254,253,251,247,239,223,191,127]
        #Mak used to put zero ex:254->11111110, 253->11111101 .. associated with AND bitwise
        self.maskZERO = self.maskZEROValues.pop(0)
        
        self.curwidth = 0  # Current width position
        self.curheight = 0 # Current height position
        self.curchan = 0   # Current channel position

    @staticmethod
    def is_inputs_valid(data) -> bool:
        "Check if inputs are valid."
        l = len(data)
        if self.width*self.height*self.nbchannels < l+64:
            raise LSBLSBSteganographyException("Carrier image not big enough to hold all the datas to steganography")

    def put_binary_value(self, bits): #Put the bits in the image
        for c in bits:
            val = list(self.image[self.curheight,self.curwidth]) #Get the pixel value as a list
            if int(c) == 1:
                val[self.curchan] = int(val[self.curchan]) | self.maskONE #OR with maskONE
            else:
                val[self.curchan] = int(val[self.curchan]) & self.maskZERO #AND with maskZERO
                
            self.image[self.curheight,self.curwidth] = tuple(val)
            self.next_slot() #Move "cursor" to the next space
        
    def next_slot(self):#Move to the next slot were information can be taken or put
        if self.curchan == self.nbchannels-1: #Next Space is the following channel
            self.curchan = 0
            if self.curwidth == self.width-1: #Or the first channel of the next pixel of the same line
                self.curwidth = 0
                if self.curheight == self.height-1:#Or the first channel of the first pixel of the next line
                    self.curheight = 0
                    if self.maskONE == 128: #Mask 1000000, so the last mask
                        raise LSBSteganographyException("No available slot remaining (image filled)")
                    else: #Or instead of using the first bit start using the second and so on..
                        self.maskONE = self.maskONEValues.pop(0)
                        self.maskZERO = self.maskZEROValues.pop(0)
                else:
                    self.curheight +=1
            else:
                self.curwidth +=1
        else:
            self.curchan +=1

    def read_bit(self): #Read a single bit int the image
        val = self.image[self.curheight,self.curwidth][self.curchan]
        val = int(val) & self.maskONE
        self.next_slot()
        if val > 0:
            return "1"
        else:
            return "0"
    
    def read_byte(self):
        return self.read_bits(8)
    
    def read_bits(self, nb): #Read the given number of bits
        bits = ""
        for i in range(nb):
            bits += self.read_bit()
        return bits

    def byteValue(self, val):
        return self.binary_value(val, 8)
        
    def binary_value(self, val, bitsize): #Return the binary value of an int as a byte
        binval = bin(val)[2:]
        if len(binval) > bitsize:
            raise LSBSteganographyException("binary value larger than the expected size")
        while len(binval) < bitsize:
            binval = "0"+binval
        return binval

    def encode_text(self, txt):
        l = len(txt)
        binl = self.binary_value(l, 16) #Length coded on 2 bytes so the text size can be up to 65536 bytes long
        self.put_binary_value(binl) #Put text length coded on 4 bytes
        for char in txt: #And put all the chars
            c = ord(char)
            self.put_binary_value(self.byteValue(c))
        return self.image
       
    def decode_text(self):
        ls = self.read_bits(16) #Read the text size in bytes
        l = int(ls,2)
        i = 0
        unhideTxt = ""
        while i < l: #Read all bytes of the text
            tmp = self.read_byte() #So one byte
            i += 1
            unhideTxt += chr(int(tmp,2)) #Every chars concatenated to str
        return unhideTxt

    def encode_image(self, imtohide):
        w = imtohide.width
        h = imtohide.height
        if self.width*self.height*self.nbchannels < w*h*imtohide.channels:
            raise LSBSteganographyException("Carrier image not big enough to hold all the datas to steganography")
        binw = self.binary_value(w, 16) #Width coded on to byte so width up to 65536
        binh = self.binary_value(h, 16)
        self.put_binary_value(binw) #Put width
        self.put_binary_value(binh) #Put height
        for h in range(imtohide.height): #Iterate the hole image to put every pixel values
            for w in range(imtohide.width):
                for chan in range(imtohide.channels):
                    val = imtohide[h,w][chan]
                    self.put_binary_value(self.byteValue(int(val)))
        return self.image

                    
    def decode_image(self):
        width = int(self.read_bits(16),2) #Read 16bits and convert it in int
        height = int(self.read_bits(16),2)
        unhideimg = np.zeros((width,height, 3), np.uint8) #Create an image in which we will put all the pixels read
        for h in range(height):
            for w in range(width):
                for chan in range(unhideimg.channels):
                    val = list(unhideimg[h,w])
                    val[chan] = int(self.read_byte(),2) #Read the value
                    unhideimg[h,w] = tuple(val)
        return unhideimg
    
    def encode_binary(self, data):
        l = len(data)
        if self.width*self.height*self.nbchannels < l+64:
            raise LSBSteganographyException("Carrier image not big enough to hold all the datas to steganography")
        self.put_binary_value(self.binary_value(l, 64))
        for byte in data:
            byte = byte if isinstance(byte, int) else ord(byte) # Compat py2/py3
            self.put_binary_value(self.byteValue(byte))
        return self.image

    def decode_binary(self):
        l = int(self.read_bits(64), 2)
        output = b""
        for i in range(l):
            output += bytearray([int(self.read_byte(),2)])
        return output


class Compare():
    def __init__(self, img1, img2):
        self.img1 = img1
        self.img2 = img2
    def get_results(self):
        print("meanSquareError: %s" %(self.meanSquareError()))
        print("psnr: %s" %(self.psnr()))
    def correlation(self):
        return signal.correlate2d (self.img1, self.img2)
    def meanSquareError(self):
        error = np.sum((self.img1.astype('float') - self.img2.astype('float')) ** 2)
        error /= float(self.img1.shape[0] * self.img1.shape[1]);
        return error
    def psnr(self):
        mse = self.meanSquareError()
        if mse == 0:
            return 100
        PIXEL_MAX = 255.0
        return 20 * math.log10(PIXEL_MAX / math.sqrt(mse))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['steganography_tools']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8.2.0,<9.0.0', 'numpy>=1.18,<1.20']

entry_points = \
{'console_scripts': ['cli_command_hello = steganography_tools:main']}

setup_kwargs = {
    'name': 'steganography-tools',
    'version': '0.1.3',
    'description': 'A Python library for higher level python functions used in image analysis and computer vision for steganography',
    'long_description': '# Steganography Tools\n\n**Authors:** Sébastien Mascha & Thomas Le Couédic\n\n## Installation\n\n`pip install steganography-tools`\n\n## References\n\n### Module: Steganography for Image, Sounds, Text\n\n**Available formats**\n- PNG\n- JPEG\n\n**Importation**\n\n`from steganography_tools import st`\n\n**Usage**\n\n```\n# Encoding\nsteg = st.LSBSteganography(cv2.imread("image.jpg"))\nimg_encoded = steg.encode_text("my message")\ncv2.imwrite("image_enc.png", img_encoded)\nplt.imshow(img_encoded)\n\n# Decoding\nim = cv2.imread("image_enc.png")\nsteg = st.LSBSteganography(im)\nprint("Text value:",steg.decode_text())\n```\n\n**Compare original and encoded images**\n\n```\noriginal = cv2.imread(\'image.jpg\')\nlsbEncoded = cv2.imread("image_enc.png")\noriginal = cv2.cvtColor(original, cv2.COLOR_BGR2RGB)\nlsb_encoded_img = cv2.cvtColor(lsbEncoded, cv2.COLOR_BGR2RGB)\n\ncompare_images = st.Compare(original, lsb_encoded_img)\ncompare_images.get_results()\n```\n\n### Module: Image processing\n\n**Importation**\n\n`from steganography_tools import image_processing as st_processing`\n\n**Usage**\n\n```\nTEST_PHOTO = "image.jpg"\n\n# This function allows us to get the basic information of an image\nst_processing.img_information(TEST_PHOTO)\n\n# Check for potential metadata\nst_processing.get_metadata(TEST_PHOTO)\n\n# TRANSFORMATION\n\nst_processing.rgb2grayscale(TEST_PHOTO_GRAYSCALE)\n\nTEST_PHOTO = cv2.imread(TEST_PHOTO) \nprint("Type: ",type(TEST_PHOTO))\nTEST_PHOTO = TEST_PHOTO[:,:,0]\n\nst_processing.plot_histogram(TEST_PHOTO)\nst_processing.thresholding(TEST_PHOTO, 55)\n\n# Left to right : grayscale image | contrast increased\nst_processing.display_images(TEST_PHOTO)\n\n# Image manipulation and numpy arrays\nst_processing.image_manipulation("image_grayscale.jpg")\n\n# Geometrical transformations\nst_processing.geo_transfomation(TEST_PHOTO)\n\n# Blurring\nst_processing.blurring(TEST_PHOTO)\n\n# Sharpenning\nst_processing.sharpenning(TEST_PHOTO)\n\n# Denoising\nst_processing.denoising(TEST_PHOTO)\n```\n\n## CLI\n\n```\nCommand Line Arguments:\n -h, --hide                      To hide data in an image file\n -r, --recover                   To recover data from an image file\n -i, --input TEXT                Path to an bitmap (.bmp or .png) image\n -s, --secret TEXT               Path to a file to hide in the image\n -o, --output TEXT               Path to an output file\n --help                          Show this message and exit.\n```',
    'author': 'Sebastien Mascha',
    'author_email': 'sebastien.mascha@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sebastienmascha/data-science-package-steganography-tools',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
